#!/usr/pkg/bin/perl

use Device::SerialPort;
use Fcntl;
use Getopt::Long;

my $device;
my $precount = 8;
my $count_before_exit = 15;
my $send_command;
my $command_count = 4;
my @default_commands = ("a0a1000405000500000d0a","a0a10009080100000000000000090d0a","a0a100030e0a00040d0a");
my @commands;
my $baud_rate = 9600;
my $find_baudrate;
my $find_baud_try = 10;
my $debug;
my %baudmapping = (0 => 4800,
		   1 => 9600,
		   2 => 19200,
		   3 => 38400,
		   4 => 57600,
		   5 => 115200);
my $reading_timeout = 15;

GetOptions ("device=s" => \$device,
	    "precount=i"   => \$precount,
	    "count-before-exit|count_before_exit=i" => \$count_before_exit,
	    "send-command|send_command" => \$send_command,
	    "command-count|command_count=i" => \$command_count,
	    "command=s" => \@commands,
	    "baud-rate|baud_rate=i" => \$baud_rate,
	    "find-baud-rate|find-baudrate|find_baudrate" => \$find_baudrate,
	    "find-baudrate-tries|find_baudrate_tries" => \$find_baud_try,
	    "debug+" => \$debug,
	    "read-timeout|read_timeout=i" => \$reading_timeout
    )
    or die("Error in command line arguments\n");

if (defined $device) {
    my $port = new Device::SerialPort($device);
    die("Could not open port\n") if ! defined $port;
    $port->user_msg(ON); 
    $port->baudrate($baud_rate); 
    $port->parity("none"); 
    $port->databits(8); 
    $port->stopbits(1); 
    $port->handshake("none"); 
    $port->write_settings;

    $port->lookclear; 

    if (defined $find_baudrate) {
	my @hunt = (9600,115200,4800,19200,38400,57600);
	my $thisbaud;
	my $trycount = 1;
	
	do {
	    foreach (@hunt) {
		my $tcount;
		my $xcount;
		my $count;
		my $saw;
		my @acc = ();
		my $nextbaud;
		my $zerocount = 0;
		my $xtime;

		$port->lookclear;
		$port->baudrate($_);
		$port->write_settings;
		$port->lookclear;

		print "Trying baud rate: $_\n";

		$tcount = 0;
		$xcount = 0;
		$xtime = time;
		do {
		    my $ztime = time;
		    ($count,$saw) = $port->read(100);
		    if ($count > 0) {
			$xtime = time;
		    }
		    $tcount += $count;
		    if (defined $debug && $debug > 3) {
			print "FLUSH $count $tcount $xtime $ztime\n";
		    }
		    if (($ztime - $xtime) > $reading_timeout) {
			if (defined $debug && $debug > 2) {
			    print "FLUSH EXIT $count $tcount $xtime $ztime\n";
			}
			$xcount++;
		    }
		} until ($tcount > 80 || $xcount > 0);
		$port->lookclear;

		if ($xcount > 0) {
		    if (defined $debug && $debug > 2) {
			print "FLUSH TIMEOUT, NEXT BAUD\n";
		    }
		    next;
		}

		$xtime = time;
		while (1) {
		    ($count,$saw) = $port->read(1);
		    if ($count > 0) {
			$xtime = time;
			my $c = ord $saw;
			if (defined $debug && $debug > 2) {
			    print "CHECKING WITH: $c ";
			    if ($c >= 32 && $c < 127) {
				print "-$saw-\n";
			    } else {
				print "\n";
			    }
			}
			if ($c == 0) {
			    if (defined $debug && $debug > 2) {
				print "C IS ZERO\n";
			    }
			    $zerocount++;
			    if ($zerocount == 6) {
				if (defined $debug && $debug > 2) {
				    print "C IS ZERO, COUNT EXCEEDED, NEXT BAUD\n";
				}
				$nextbaud++;
				last;
			    } else {
				if (defined $debug && $debug > 2) {
				    print "C IS ZERO, COUNT OK, NEXT CHAR\n";
				}
				next;
			    }
			} else {
			    $zerocount = 0;
			}
			if ($c > 127) {
			    if (defined $debug && $debug > 2) {
				print "C TOO BIG, NEXT BAUD\n";
			    }
			    $nextbaud++;
			    last;
			}
			if ($c == 36) {
			    if (defined $debug && $debug > 2) {
				print "C IS DOLLAR\n";
			    }
			    if ($#acc < 0) {
				if (defined $debug && $debug > 2) {
				    print "C IS DOLLAR ACC EMPTY, NEXT CHAR\n";
				}
				@acc = ();
				next;
			    }
			    if ($#acc == 1) {
				if (defined $debug && $debug > 2) {
				    print "C IS DOLLAR ACC IS 1, THIS BAUD\n";
				}
				$thisbaud = $_;
				last;
			    }
			}
			if ($c == 13) {
			    if (defined $debug && $debug > 2) {
				print "C IS 13\n";
			    }
			    if ($#acc < 0) {
				if (defined $debug && $debug > 2) {
				    print "C IS 13 ACC EMPTY, RECORD, NEXT CHAR\n";
				}
				push(@acc,$c);
				next;
			    }
			    if ($#acc >= 0) {
				if (defined $debug && $debug > 2) {
				    print "C IS 13 ACC GE 0, NEXT BAUD\n";
				}
				$nextbaud++;
				last;
			    }
			}
			if ($c == 10) {
			    if (defined $debug && $debug > 2) {
				print "C IS 10\n";
			    }
			    if ($#acc < 0) {
				if (defined $debug && $debug > 2) {
				    print "C IS 10 ACC EMPTY, NEXT CHAR\n";
				}
				@acc = ();
				next;
			    }
			    if ($#acc == 0) {
				if (defined $debug && $debug > 2) {
				    print "C IS 10, ACC IS 0, RECORD, NEXT CHAR\n";
				}
				push(@acc,$c);
				next;
			    }
			    if ($#acc >= 0) {
				if (defined $debug && $debug > 2) {
				    print "C IS 10, ACC GE 0, NEXT BAUD\n";
				}
				$nextbaud++;
				last;
			    }
			}
			if ($c < 32) {
			    if (defined $debug && $debug > 2) {
				print "C TOO SMALL, NEXT BAUD\n";
			    }
			    $nextbaud++;
			    last;
			}
		    } else {
			my $ztime = time;
			if (defined $debug && $debug > 3) {
			    my $qq = $ztime - $xtime;
			    print "LOOP COUNT $xtime $ztime $qq\n";
			}
			if (($ztime - $xtime) > $reading_timeout) {
			    if (defined $debug && $debug > 2) {
				print "LOOP TIMEOUT, NEXT BAUD\n";
			    }
			    $nextbaud++;
			    last;
			}
 		    }
		}
		if (defined $nextbaud) {
		    if (defined $debug && $debug > 2) {
			print "GOING TO NEXT BAUD\n";
		    }
		    undef $nextbaud;
		}
		if (defined $thisbaud) {
		    if (defined $debug && $debug > 2) {
			print "USING THIS BAUD\n";
		    }
		    last;
		}
	    }

	    if (defined $thisbaud) {
		print "Baud Is: $thisbaud\n";
	    } else {
		print "Could not find baud rate, try $trycount for max $find_baud_try\n";
	    }

	    $trycount++;
	} until (defined $thisbaud || $trycount == $find_baud_try);
    }

    if ($#commands == -1) {
	@commands = @default_commands
    }

    my $line;
    while ($precount > 0) {
	$line = get_a_line($port);
	print $line;
	$precount--;
    }

    if (defined $send_command) {
	my $cmd;
	foreach $cmd (@commands) {
	    my $n = $command_count;
	    if (defined $debug && $debug > 0) {
		print "SENDING COMMAND: $cmd\n";
	    }
	    $hxcommand = pack "H*", $cmd;
	    $port->write($hxcommand);
	    if ($cmd =~ /^a0a1000405000([012345])/i ||
		$cmd =~ /^a0a1000204/i) {
		my $bindex;
		if ($cmd =~ /^a0a1000204/i) {
		    $bindex = 1;
		} else {
		    $bindex = $1;
		}
		if (defined $baudmapping{$bindex}) {
		    my $mm = $baudmapping{$bindex};
		    my $xcount = 0;
		    print "Switching baudrate to $mm\n";
		    while ($xcount < 10) {
			$line = get_a_line($port);
			print $line;
			$xcount++;
			if ($line =~ /^.ACK 0[45]/) {
			    if (defined $debug && $debug > 1) {
				print "ACTUALLY SWITCHING BAUDRATE\n";
			    }
			    $port->baudrate($mm);
			    $port->write_settings;
			    $port->lookclear;
			    last;
			}
			if ($line =~ /^.NACK 0[45]/) {
			    last;
			}
		    }
		    if (defined $debug && $debug > 1) {
			print "BAUD RATE READ BACK $xcount\n";
		    }
		}
	    }
	    while ($n > 0 && $#commands > 0) {
		$line = get_a_line($port);
		print $line;
		$n--;
	    }
	}
    }

    while ($count_before_exit > 0) {
	$line = get_a_line($port);
	print $line;
	$count_before_exit--;
    }
    
  OUT:

    undef $port;
} else {
    print "Missing device\n";
}

sub get_a_line {
    my $p = shift;
    my $line = "";
    my @line = ();

    while (1) {
	my ($count,$saw)=$p->read(1);
	if ($count > 0) {
	    push(@line,$saw);
	    $hex = unpack "H*", $saw;
	    if ($hex eq "0a") {
		if (($#line - 1) > -1) {
		    $hex0 = unpack "H*", $line[$#line - 1];
		    if ($hex0 eq "0d") {
			$line = join('',@line);
			if ($line =~ /^\$/) {
			    last;
			} else {
			    $hex = unpack "H*", $line;
			    if ($hex =~ /^a0a1/) {
				if ($hex eq "a0a100028300830d0a") {
				    $line = "[COMMAND ACK]\n";
				    last;
				} else {
				    if ($hex =~ /^a0a1000283([0-9a-f][0-9a-f])[0-9a-f][0-9a-f]0d0a$/) {
					$line = "[ACK $1]\n";
					last;
				    } else {
					if ($hex =~ /^a0a1000284([0-9a-f][0-9a-f])[0-9a-f][0-9a-f]0d0a$/) {
					    $line = "[NACK $1]\n";
					} else {
					    $line = "$hex\n";
					    last;
					}
				    }
				}
			    } else {
				last;
			    }
			}
		    }
		} else {
		    $line = join('',@line);
		}
	    }
	}
    }

    return $line;
}
